#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 27 21:07:29 2018

@author: loku
"""


from rdkit import Chem
import time
cur_time = time.asctime(time.localtime(time.time()))
print("version_12_2.py"+str(cur_time))
double1 = {'L': 'Cl', 'B': 'Br', 'D': '[C@@H]', 'E': '[C@H]', 'Z': '[nH]', 'Y': '[N+]', 'X': '[O-]', 'W': '[n+]', 'V': '[N-]'}
double2 = {'Cl': 'L', 'Br': 'B', '[C@@H]': 'D', '[C@H]': 'E', '[nH]': 'Z', '[N+]': 'Y', '[O-]': 'X', '[n+]': 'W', '[N-]': 'V'}
unvisit = ["a","g","h","j","m","q","r","t","u","A","G","H","J","M","Q","R","T","U"]
ring_stru = ["useless","!","%","&","*","+",",",";","<",">"]
class fragmentNode:
    def __init__(self,x,y,z,n):
        self.atomlist = x
        self.bondlist = y
        self.notation = z
        self.nei_atom = n
        #self.visited = False

def filter_3rings(mol):
    ringInfo = mol.GetRingInfo()
    atom_in_ring = ringInfo.AtomRings()
    for ring in atom_in_ring :
        for atom in ring:
            num_rings = ringInfo.NumAtomRings(atom)
            if num_rings > 2:
                return True
    return False
    
def find_fragment(cur_idx, mol):
    stack = []
    pare = []
    stack.append(cur_idx)
    bondlist = []#visited
    atomlist = []
    nei_atom = []
    nei_bond = []
    notation = ""
    sublist = []
    count = {}
    while stack :#DFS 
        #cur_idx = stack.top()[0]忘了这个想法
        cur_idx = stack[-1]
        flag = False
        
        atom = mol.GetAtomWithIdx(cur_idx)
        ringInfo = mol.GetRingInfo()
        bonds = [bond.GetIdx() for bond in atom.GetBonds()]#bond type
        num = ringInfo.NumAtomRings(cur_idx)
        #nei_bond include all nei bonds of this fragment
        bonds = list(set(bonds) - set(bondlist) - set(nei_bond))
        #bonds = list(set(bonds) - set(bondlist))
        if cur_idx in atomlist:#判断当前点是不是访问过
            flag = True
            #n = count[cur_idx]
        else:#当前点没有被访问过
            atomlist.append(cur_idx)
            for b in bonds:
                bond = mol.GetBondWithIdx(b)
                atom2 = bond.GetOtherAtom(atom)
                idx = atom2.GetIdx()
                if idx in atomlist:
                    bondlist.append(b)
                    n = count[idx]
                    count[idx] = n - 1
                    pare.append((cur_idx,idx))
                    continue
                    #这里可以获取环破的地方
                num2 = ringInfo.NumAtomRings(idx)
                inR = bond.IsInRing()
                if inR or not inR and num2 == 0 and num == 0:
                    #判断邻边是不是和当前点在同一个结构里面
                    continue
                else:
                    nei_bond.append(b)
                    nei_atom.append(idx)
            bonds = list(set(bonds) - set(bondlist) - set(nei_bond))
            count[cur_idx] = len(bonds)#number of nei_bon of cur_atom
        a_s = atom.GetSmarts()
        
        new =[]
        if len(a_s) > 1:#对【c@】这种符号进行处理
            new.append(a_s)
            if a_s in double2.keys():
                a_s = double2[a_s]
            else:
                u = unvisit.pop()
                double2[a_s] = u
                double1[u] = a_s
        else:
            pass
        if bonds and not flag:
            for b in bonds :
                bond = mol.GetBondWithIdx(b)
                atom2 = bond.GetOtherAtom(atom)
                idx = atom2.GetIdx()
                stack.append(b)
                bondlist.append(b)
                stack.append(idx) 
                break
        elif bonds and flag:
            sublist.append("(" + notation + ")")
            notation = ""
            for b in bonds :
                bond = mol.GetBondWithIdx(b)
                atom2 = bond.GetOtherAtom(atom)
                idx = atom2.GetIdx()
                stack.append(b)
                bondlist.append(b)
                stack.append(idx)
                break
        elif not bonds and not flag:
            
            if len(stack) == 1:#come to the bot
                notation = a_s + notation
                stack.pop()
            else:
                notation = a_s
                stack.pop()
                b = stack.pop()#get bond info
                bond = mol.GetBondWithIdx(b)
                #这个bond和atom的顺序对吗？
                ring_p = ""
                #break_b_symbol = ""
                for i,p in enumerate(pare):
                    if cur_idx in p:
                        #ring_p += str(i+1)
                        break_b = mol.GetBondBetweenAtoms(p[0],p[1])
                        break_b_symbol = break_b.GetSmarts()
                        ring_p += break_b_symbol
                        ring_p += ring_stru[i+1]
                #notation =  bond.GetSmarts() + a_s + break_b_symbol + ring_p
                notation =  bond.GetSmarts() + a_s + ring_p
                #sublist.append( notation )
                #notation = ""
        else:
            s = ""
            x = count[cur_idx]
            #if x > 1:#这样做是因为不经过分岔点的没有放入sub里面
            #现在不经过的也放到sub里面
            for i  in range (x -1 ):#did not append last sub
#                sub = sublist[0]
#                s += sub
#                if len(sublist) - 1  != 0:
#                    sublist = sublist[1:]
#                else:
#                    sublist = []
                #这里要确保每次选择的子集是属于当前点的，所以从后往前加
                sub = sublist.pop()
                #但是会出现atomlist里面排前面的atom符号出现在了后面
                #因此这里是s = sub + s
                s = sub + s
            if len(stack) == 1:
                ring_p = ""
                stack.pop()
                for i,p in enumerate(pare):
                    if cur_idx in p:
                        #ring_p += str(i+1)
                        ring_p += ring_stru[i+1]
                notation = a_s + ring_p + s + notation
                #sublist.append("(" + notation)
            else:
                stack.pop()
                b = stack.pop()
                bond = mol.GetBondWithIdx(b)
                ring_p = ""
                #break_b_symbol = ""
                for i,p in enumerate(pare):
                    if cur_idx in p:
                        #ring_p += str(i+1)
                        break_b = mol.GetBondBetweenAtoms(p[0],p[1])
                        break_b_symbol = break_b.GetSmarts()
                        ring_p += break_b_symbol
                        ring_p += ring_stru[i+1]
                #notation = bond.GetSmarts() + a_s + break_b_symbol + ring_p + s + notation
                notation = bond.GetSmarts() + a_s + ring_p + s + notation
                #sublist.append("(" + notation + ")")
    fragment = fragmentNode(atomlist,bondlist,notation,nei_atom)
    return fragment

def formA_(idx,mol,visited):#nei_pare最好存一对，这样还方便确认bond
    fragment = find_fragment(idx, mol)
    s = fragment.notation
    visited += fragment.atomlist
    nei = fragment.nei_atom
    nei = list(set(nei) - set(visited))
    if nei:
        sublist = ""
        non_c_sublist  = ""
        bond_symbol = ""
        for n in nei:
            atomlist = fragment.atomlist
            atom = mol.GetAtomWithIdx(n)
            for a in atom.GetNeighbors():
                i = a.GetIdx()
                if i in atomlist:
                    bond = mol.GetBondBetweenAtoms(i,n)
                    bond_symbol = bond.GetSmarts()
                    break
            p = atomlist.index(i)
            notation, non_c_notation,visited = formA_(n,mol,visited)
            sublist = sublist + "{" + str(p) + bond_symbol + notation + "}"
            non_c_sublist =  non_c_sublist + "{" + bond_symbol + notation + "}"
        return fragment.notation + sublist, fragment.notation + non_c_sublist,visited
            
    else:
        return fragment.notation, fragment.notation, visited


def encode(smi):
    #print(smi)
    mol = Chem.MolFromSmiles(smi)
#    right = []
#    left = []
#    
#    for s in smi:
#        if s == "\\":
#            
#        elif s == "/"
#    
    #step 1: use filter remove 3_rings
    if filter_3rings(mol):
        return "false","false"
    visited = []
    formA, non_c_formA, visited = formA_(0,mol,visited)
    return formA
#print(encode_formA("CN1CCN(CC2=c3oc(=O)/c(=C4\C=CC=CN4)cc3C=CC2=O)CC1"))
    
#import pickle
#
#f = open("smiles_validAll.txt","rb")
#smiles = pickle.load(f)
#f.close()
##smiles = "O=C(CN(Cc1ccc(F)cc1)C(=O)CCC(=O)Nc1ccccn1)NCc1ccco1"
#count = 0
#i = 0
#bsmiles = []
#for s in smiles:
#    i += 1
#    #print("smiles" + str(i) + ":  ",s)
##    !!
#    formA,non_c_formA = encode_formA(s)
#    if formA == "false":
#        continue
#    else:
#        bsmiles.append(formA)
#f = open("bsmiles_valid","wb")
#pickle.dump(bsmiles,f)
#f.close()

    







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan  8 16:27:14 2019

@author: xiaochun

"""

"""
convert bsmiles back to smiles
call function "convertB"
"""
ring_stru = ["useless","!","%","&","*","+",",",";","<",">"]
#ring_stru = ["useless","!","$","%","&","*","+",",","-","."]
def combin(parent,child,dic):
    atom_mark = ["A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z"]
    atom_mark2 = ["a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z"]
    mark = atom_mark + atom_mark2
    position = ""
    for s in child:
        if not s.isdigit():
            break
        position += s
    l = len(position)
    position = int(position)
    count = -1
    for i,s in enumerate(parent):
        if s.isdigit() or s not in mark :
            continue
        else:
            count += 1
            if count == position:
                break
    idx = len(dic)
    key = "!" + str(idx) + "!"
    dic[key] = child[l:]
    parent = parent[:i + 1] + "(" + key + ")" +parent[i + 1:]
    return parent,dic            
    
def insert_sub(smi,dic):
    flag = False
    smi2 = ""
    idx = ""
    start = False
    for s in smi:
        if s == "!" and len(idx) == 0:
            start = True
            continue
        elif s == "!" and len(idx)!= 0:
            start = False
            idx = int(idx)
            idx = "!" + str(idx) + "!"
            smi2 += dic[idx]
            idx = ""
            flag = True
            continue
        if start == False:
            smi2 += s
            continue
        idx += s
    return smi2,flag

def change_no(content,ring_p):
    smi = ""
    visited = {}
    position = ""
    for s in content:
        if not s.isdigit():
            break
        position += s
    l = len(position)
    content = content[l:]
    for s in content:
        if s.isdigit() and int(s) not in visited.keys():
            visited[int(s)] = str(ring_p)
            smi += str(ring_p)
            ring_p += 1
            continue
        elif s.isdigit() and int(s) in visited.keys():
            
            smi += visited[int(s)]
            continue
        smi += s
    return position + smi,ring_p

def convertB(bsmiles):
    #bsmiles:string
    double1 = {'L': 'Cl', 'B': 'Br', 'D': '[C@@H]', 'E': '[C@H]', 'Z': '[nH]', 'Y': '[N+]', 'X': '[O-]', 'W': '[n+]', 'V': '[N-]'}
#    double2 = {'Cl': 'L', 'Br': 'B', '[C@@H]': 'D', '[C@H]': 'E', '[nH]': 'Z', '[N+]': 'Y', '[O-]': 'X', '[n+]': 'W', '[N-]': 'V'}
#    unvisit = ["a","g","h","j","m","q","r","t","u","A","G","H","J","M","Q","R","T","U"]
    
    double = ['L', 'B', 'D', 'E', 'Z', 'Y', 'X', 'W', 'V']
    lev_no = 0
    level_list = [[0]]
    content = ""
    dic = {}
    smi = ""
    for s in bsmiles:
        if s in ring_stru:
            i = ring_stru.index(s)
            smi += str(i)
            continue
        
        smi += s
    ring_p = 1
    for s in smi:
        if s == "{":
            if len(content) != 0 and len(level_list) > lev_no:
                content,ring_p = change_no(content,ring_p)
                level_list[lev_no].append(content)
            elif len(content) != 0:
                content,ring_p = change_no(content,ring_p)
                level_list.append([content])
            lev_no += 1
            content = ""
            continue
        elif s == "}":
            if len(content) != 0:
                content,ring_p = change_no(content,ring_p)
                parent = level_list[lev_no - 1].pop()
                parent,dic = combin(parent,content,dic)
                level_list[lev_no - 1].append(parent)
            else:
                parent = level_list[lev_no - 1].pop()
                child = level_list[lev_no].pop()
                parent,dic = combin(parent,child,dic)
                level_list[lev_no - 1 ].append(parent)
            lev_no = lev_no - 1
            content = ""
            continue
        content += s
    if len(content) != 0:#it means no {} in this string
        level_list[0].append(content)
        
    smi = level_list[0]
    smi = smi[1]
    
    #location = []
    flag = True
    while flag:
        smi,flag = insert_sub(smi,dic)
    
    #add ring number and [c@@h] cl br 
    smi2 = ""
    for s in smi:
        if s in double:
            i = double1[s]
            smi2 += i
            continue
        smi2 += s
    
    return smi2        

"""
seperate bsmiels into fragments
"""
import pickle
from rdkit import Chem

def getFragments(visited,bsmiles):
    #visited:list
    #bsmiles:string
    content =""
    fragments= []
    for s in bsmiles:
        if s == "{":
            position = ""
            if len(content) != 0:
                for c in content:
                    if c.isdigit():
                        position += c
                    else:
                        break
                l = len(position)
                fragments.append(position)
                if content[l:] not in visited:
                    visited.append(content[l:])
                fragments.append(content[l:]) 
                
                content = ""
            
            fragments.append("{")
                
            continue
        elif s == "}":
            position = ""
            if len(content) != 0:
                for c in content:
                    if c.isdigit():
                        position += c
                    else:
                        break
                l = len(position)
                fragments.append(position)
                if content[l:] not in visited:
                    visited.append(content[l:])
                fragments.append(content[l:]) 
                content = ""
            fragments.append("}")
                
            continue
        content += s
    if len(content ) != 0:
        if content not in visited:
            visited.append(content)
        fragments.append(content)
    return visited,fragments



def bsmiles_to_frag(inputfile,outputfile):
    #inputfile:filenam(string)
    #outputfile:filename(string)
    f = open(inputfile,"rb")
    bsmiles = pickle.load(f)
    f.close()
    fragments = []
    i = 0
    visited = []
    for s in bsmiles:
        #print("bsmiles:          ",s)
        i += 1
        #s = "CC(=O)N{3c!scnc!{4-c!cccs!}{2NC=O{1c!ccccc!}}}"
        visited,frag = getFragments(visited,s)#visited 是用来测试一共有多少个fragment
        fragments.append(frag)

    f = open(outputfile,"wb")
    pickle.dump(fragments,f)
    f.close()
    
"""
make sure that every bsmiles seperated into correct fragments.
"""
def result_test(fragments,bsmiles):
    #fragments: [fragment:string,fragment:string,...]
    #bsmiles:string
    ss = ""
    for j in fragments:
        ss += j
    if ss != bsmiles:
        return False
    try:
        smiles = convertB(ss)
        #print("smiles1:           ",smiles)
        mol = Chem.MolFromSmiles(smiles)
        smile2 = Chem.MolToSmiles(mol)
    except:
        return False
    return True
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 21 15:04:11 2019

@author: lu
"""

from setuptools import setup,find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name="BSMILES",
      version="0.0.1",
      author="lu",
      author_email="lu@gmail.com",
      description="transformation between smiles and bsmiles",
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=find_packages(),
      url = "https://github.com/LUJUANJUAN/BSMILES",
      classifiers=[
              "Programming Language :: Python :: 3",
              "License :: OSI Approved :: MIT License",
              "Operating System :: OS Independent",
              ],)